/****************************************************************************

 PowerUnit
		Copyright(C) Denshi Usagi

****************************************************************************/

// PIC24FV32KA302 Configuration Bit Settings
// FBS
#pragma config BWRP = OFF               // Boot Segment Write Protect (Disabled)
#pragma config BSS = OFF                // Boot segment Protect (No boot program flash segment)

// FGS
#pragma config GWRP = OFF               // General Segment Write Protect (General segment may be written)
#pragma config GSS0 = OFF               // General Segment Code Protect (No Protection)

// FOSCSEL
#pragma config FNOSC = FRCDIV           // Oscillator Select (8MHz FRC oscillator With Postscaler (FRCDIV))
#pragma config SOSCSRC = DIG            // SOSC Source Type (Digital Mode for use with external source)
#pragma config LPRCSEL = LP             // LPRC Oscillator Power and Accuracy (Low Power, Low Accuracy Mode)
#pragma config IESO = ON                // Internal External Switch Over bit (Internal External Switchover mode enabled (Two-speed Start-up enabled))

// FOSC
#pragma config POSCMOD = NONE           // Primary Oscillator Configuration bits (Primary oscillator disabled)
#pragma config OSCIOFNC = OFF           // CLKO Enable Configuration bit (CLKO output disabled)
#pragma config POSCFREQ = HS            // Primary Oscillator Frequency Range Configuration bits (Primary oscillator/external clock input frequency greater than 8MHz)
#pragma config SOSCSEL = SOSCLP         // SOSC Power Selection Configuration bits (Secondary Oscillator configured for low-power operation)
#pragma config FCKSM = CSDCMD           // Clock Switching and Monitor Selection (Both Clock Switching and Fail-safe Clock Monitor are disabled)

// FWDT
#pragma config WDTPS = PS8192           // Watchdog Timer Postscale Select bits (1:8192)
#pragma config FWPSA = PR32             // WDT Prescaler bit (WDT prescaler ratio of 1:32)
#pragma config FWDTEN = SWON            // Watchdog Timer Enable bits (WDT controlled with the SWDTEN bit setting)
#pragma config WINDIS = OFF             // Windowed Watchdog Timer Disable bit (Standard WDT selected(windowed WDT disabled))

// FPOR
#pragma config BOREN = BOR3             // Brown-out Reset Enable bits (Brown-out Reset enabled in hardware, SBOREN bit disabled)
#pragma config LVRCFG = OFF             // Low Voltage Regulator Configuration bit (Low Voltage regulator is not available)
#pragma config PWRTEN = ON              // Power-up Timer Enable bit (PWRT enabled)
#pragma config I2C1SEL = PRI            // Alternate I2C1 Pin Mapping bit (Use Default SCL1/SDA1 Pins For I2C1)
#pragma config BORV = V20               // Brown-out Reset Voltage bits (Brown-out Reset set to lowest voltage (2.0V))
#pragma config MCLRE = ON               // MCLR Pin Enable bit (RA5 input pin disabled,MCLR pin enabled)

// FICD
#pragma config ICS = PGx1               // ICD Pin Placement Select bits (EMUC/EMUD share PGC1/PGD1)

// FDS
#pragma config DSWDTPS = DSWDTPSF       // Deep Sleep Watchdog Timer Postscale Select bits (1:2,147,483,648 (25.7 Days))
#pragma config DSWDTOSC = LPRC          // DSWDT Reference Clock Select bit (DSWDT uses Low Power RC Oscillator (LPRC))
#pragma config DSBOREN = ON             // Deep Sleep Zero-Power BOR Enable bit (Deep Sleep BOR enabled in Deep Sleep)
#pragma config DSWDTEN = ON             // Deep Sleep Watchdog Timer Enable bit (DSWDT enabled)


#include <xc.h>
#include "typedefs.h"

//=================================================================

#define LCD_RS		_LATB15
#define LCD_E		_LATA7
#define LCD_DB0		_LATB14
#define LCD_DB1		_LATB13
#define LCD_DB2		_LATB12
#define LCD_DB3		_LATB11
#define LCD_DB4		_LATB10
#define LCD_DB5		_LATB9
#define LCD_DB6		_LATB8
#define LCD_DB7		_LATB7

#define DAC_CS		_LATA4
#define RE1_A		_RB2
#define RE1_B		_RB3
#define RE2_A		_RA0
#define RE2_B		_RA1

#define ADCH_V		0b01101	// ADC-I:AN13
#define ADCH_I		0b01110	// ADC-V:AN14
#define ADCH_VBG	0b11010	// Internal Band Gap Reference (VBG)

#define EEP_FLAG		0
#define EEP_SETV		1
#define EEP_SETI		2
#define EEP_VSET_BASE	0	// 5～130
#define EEP_ISET_BASE	130	// 131～141

//=================================================================

#define MODE_NORMAL		0	// 通常動作
#define MODE_TEST0		1	// 0点チェックモード：電圧表示と電流表示が0であることを確認する
#define MODE_TEST1		2	// 測定電圧確認モード：12Vを出力、テスターと同じ電圧表示になるようにVR4を調整する
#define MODE_TEST2		3	// 電圧校正モード：電圧DAC値を自動調整する（2,3を繰り返し12V設定時に12V表示になるようにする）
#define MODE_TEST3		4	// 電流校正モード：電流DAC値を自動調整する（1V, 出力をショートさせて実行する）

#define RE_NONE			0
#define RE1_UP			1
#define RE1_DOWN		2
#define RE2_UP			3
#define RE2_DOWN		4

#define MIN_V			5	// 5～130 0.5V～13.0V (0.1V Step)
#define MAX_V			130
#define MIN_I			1	// 1～100 10mA～1A (10mA Step)
#define MAX_I			100

#define TO_TEST			5	// 電圧電流設定両方とも最小値を超えて左に回すとテストモードに入る
#define TEST_V			120 // テスト電圧
#define COMPLETE_VAL	7	// テスト完了値

//=================================================================

static void InitSystem(void);
static void Run(void);

static void InitDACVals(void);
static void SetOutput(void);
static void FineTuning(void);
static void Measurement(void);
static void Display(void);
static void FormatV(char* pBuf, WORD wVal);
static void FormatI(char* pBuf, WORD wVal);
static void DisplayText(int x, int y, const char* pText);
static int GetRE(void);
static int GetRE1(void);
static int GetRE2(void);
static void Operation(int nRE);

static void ToNormal(void);
static void ToTest(void);
static void AutoAdjV(void);
static void AutoAdjI(void);

static WORD ReadEEP(WORD wAdd);
static void WriteEEP(WORD wAdd, WORD wData);
static void InitOLED(void);
static void WriteLCD(BYTE rs, BYTE data);
static void SetDAC(WORD wA, WORD wB);
static BYTE SendSPI(BYTE bData);
static WORD GetADC(BYTE ch);

static void DelayMs(WORD wMs);
static void DelayUs(WORD wMs);

//=================================================================

static int g_nMode;
static WORD g_wSetV; // 5～130 0.5V～13.0V (0.1V Step)
static WORD g_wSetI; // 2～110 1mA～10mA～1A (1/10mA Step)
static WORD g_wBase;
static WORD g_wGetV = 0;
static WORD g_wGetI = 0;
static int g_nMinV = 0;
static int g_nMinI = 0;
static WORD g_wCurV = 0xFFFF;
static WORD g_wCurI = 0xFFFF;
static BOOL g_bAsig1 = FALSE;
static BOOL g_bAsig2 = FALSE;

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// ポート初期化
	ANSA = 0; TRISA = 0xFFFF; // All Digital input
	ANSB = 0; TRISB = 0xFFFF; // All Digital input

	LATA = 0; PORTA = 0;
	LATB = 0; PORTB = 0;
	
	// 空きピン(LOレベル固定)
	_TRISB4 = OUTPUT;
	
	// LCD
	_TRISB15 = OUTPUT;
	_TRISB14 = OUTPUT;
	_TRISB13 = OUTPUT;
	_TRISB12 = OUTPUT;
	_TRISB11 = OUTPUT;
	_TRISB10 = OUTPUT;
	_TRISA7 = OUTPUT;
	_TRISB9 = OUTPUT;
	_TRISB8 = OUTPUT;
	_TRISB7 = OUTPUT;
	
	// DAC
	DAC_CS = 1;
	_TRISA4 = OUTPUT;
	
	// SPI2 master Mode0 2MHz
	SPI2CON1bits.CKE = 1;
	SPI2CON1bits.MSTEN = 1;
	SPI2CON1bits.SPRE = 0b110; // 2:1
	SPI2CON1bits.PPRE = 0b11; // 1:1
	SPI2STATbits.SPIROV = 0;
	SPI2STATbits.SPIEN = 1;
	
	// RE
	_TRISA0 = INPUT;
	_TRISA1 = INPUT;
	_TRISB2 = INPUT;
	_TRISB3 = INPUT;

	// AN13(ADC-V)
	_ANSA2 = 1;
	_TRISA2 = INPUT;

	// AN14(ADC-I)
	_ANSA3 = 1;
	_TRISA3 = INPUT;

	AD1CON1bits.SSRC = 0b0111; // 自動変換
	AD1CON1bits.MODE12 = 1; // 12-bit A/D operation
	AD1CON2bits.PVCFG = 0b11; // 4 * Internal VBG
	AD1CON2bits.NVCFG = 0; // AVSS
	AD1CON3bits.ADRC = 1; // RC Clock
	AD1CON3bits.SAMC = 10; // 10TAD
	_CH0NA = 0b000; // AVSS
	AD1CON1bits.ADON = 1;
	
	// ADC test
	//WORD wTest = GetADC(ADCH_VBG); // wTest≒4095

	// Timer2
	T2CON = 0x8000; // TON=1
	
	DelayMs(500);
	
	// WEH001602ABPP5N00000
	InitOLED();
	DisplayText(0, 0, "= Power supply =");
	DisplayText(0, 1, "  Rady! (^^)/~~ ");
	DelayMs(2000);
}

// ●実行ループ
static void Run(void)
{
#ifndef _DEBUG
	_SWDTEN = 1;
#endif
	if (ReadEEP(EEP_FLAG) != COMPLETE_VAL) // テスト完了値
	{
		// 完成前（設定の初期化）
		InitDACVals();
		
		g_nMode = MODE_TEST0;
		g_wSetV = 0;
		g_wSetI = 0;
		WriteEEP(EEP_SETV, MIN_V);
		WriteEEP(EEP_SETI, MIN_I);
	}
	else
	{
		// 完成後（設定のロード）
		g_nMode = MODE_NORMAL;
		g_wSetV = ReadEEP(EEP_SETV);
		g_wSetI = ReadEEP(EEP_SETI);
	}

	int i = 0;
	for(;;i++)
	{
		// 出力DAC設定
		SetOutput();
		
		// 350回に一度ファインチューニング
		if ((i % 350) == 0) {
			FineTuning();
		}
		// 100回に一度表示更新
		if ((i % 100) == 0) {
			Display();
		}
		// 電圧電流計測
		Measurement();
		// 操作処理
		Operation(GetRE());
		
		ClrWdt();
	}
}

// ●DAC値の初期設定
static void InitDACVals(void)
{
	int i;

	for(i = MIN_V; i <= MAX_V; i++)
	{
		WORD wVal = (WORD)(((DWORD)i * 100LL * 4095LL + 6600) / 13200LL); // 6600は四捨五入のため
		WriteEEP(EEP_VSET_BASE + i, wVal);
	}

	for(i = MIN_I; i <= MAX_I; i++)
	{
		DWORD mA = i * 10 - 9; // 制御バイアスを引く
		WORD wVal = (WORD)((mA * 3LL * 4095LL + 1650) / 3300LL); // 1650は四捨五入のため
		WriteEEP(EEP_ISET_BASE + i, wVal);
	}
}

// ●出力設定
static void SetOutput(void)
{
	if (g_wCurV != g_wSetV)
	{
		g_wBase = (g_wSetV >= MIN_V)? ReadEEP(EEP_VSET_BASE + g_wSetV) : 0;
		SetDAC(g_wBase, 0xFFFF);
		g_wCurV = g_wSetV;
	}

	if (g_wCurI != g_wSetI)
	{
		WORD wDAC = (g_wSetI >= MIN_I)? ReadEEP(EEP_ISET_BASE + g_wSetI) : 0;
		SetDAC(0xFFFF, wDAC);
		g_wCurI = g_wSetI;
	}
}

// ●電圧ファインチューニング
static void FineTuning(void)
{
	static WORD s_wSetV = 0xFFFF;
	static int s_nAdj = 0;
	
	if ((g_nMode != MODE_NORMAL) || (g_wSetV == 0))
	{
		s_wSetV = 0xFFFF;
		return;
	}
	if (s_wSetV != g_wSetV)
	{
		s_wSetV = g_wSetV;
		s_nAdj = 0;
	}

	// 10mV単位で比較する
	int nAdj = s_nAdj;
	WORD wL = (g_wGetV + 5) / 10;
	WORD wR = g_wSetV * 10;
	if (wL < wR) {
		if (s_nAdj < 15) s_nAdj++;
	}
	if (wL > wR) {
		if (s_nAdj > -15) s_nAdj--;
	}

	if (s_nAdj != nAdj) {
		SetDAC(g_wBase + s_nAdj, 0xFFFF);
	}
}

// ●電圧電流測定
static void Measurement(void)
{
	static WORD wV[10] = {0}, wI[10] = {0};
	int i;

	if (g_nMode == MODE_NORMAL)
	{
		static int cV = 0, cI = 0;
	
		wV[cV] = GetADC(ADCH_V) * 4; // mV
		wI[cI] = GetADC(ADCH_I) / 3; // mA
		
		if (++cV >= 10) cV = 0;
		if (++cI >= 10) cI = 0;
	}
	else
	{
		for (i = 0; i < 10; i++)
		{
			wV[i] = GetADC(ADCH_V) * 4; // mV
			DelayUs(10);
		}
		for (i = 0; i < 10; i++)
		{
			wI[i] = GetADC(ADCH_I) / 3; // mA
			DelayUs(10);
		}
	}
	
	DWORD dwV = 0, dwI = 0;
	for (i = 0; i < 10; i++)
	{
		dwV += wV[i];
		dwI += wI[i];
	}
	g_wGetV = dwV / 10;
	g_wGetI = dwI / 10;

	// VR1,3,4に流れる電流を差し引く
	WORD wAdj = g_wGetV / 2000; // 2.5K//20K//20K = 2K
	g_wGetI = (g_wGetI > wAdj)? (g_wGetI - wAdj) : 0;
}

// ●表示処理
static void Display(void)
{
	static WORD s_wSetV = 0xFFFF;
	static WORD s_wGetV = 0xFFFF;
	static WORD s_wSetI = 0xFFFF;
	static WORD s_wGetI = 0xFFFF;
	static WORD s_wMode = 0xFFFF;

	// 1234567890123456
	// 12.00V: 12.01V T
	// 1000mA: 1000mA*0
	
	// Line1
	if ((s_wSetV != g_wSetV) || (s_wGetV != g_wGetV) || (s_wMode != g_nMode))
	{
		char cLine[16];

		int i = 0;
		do cLine[i] = ' '; while (i++ < 16);
	
		FormatV(&cLine[0], g_wSetV * 100);
		cLine[6] = ':';
		FormatV(&cLine[8], g_wGetV + 5);

		if (g_nMode == MODE_NORMAL)
		{
			if (((s_wGetV + 5) / 10) >= s_wSetV * 10) {
				cLine[15] = '\xAC';
			}
			else {
				cLine[15] = ' ';
			}
		}
		else {
			cLine[15] = 'T';
		}
		
		DisplayText(0, 0, cLine);
		s_wSetV = g_wSetV;
		s_wGetV = g_wGetV;
	}
	
	// Line2
	if ((s_wSetI != g_wSetI) || (s_wGetI != g_wGetI) || (s_wMode != g_nMode))
	{
		char cLine[16];

		int i = 0;
		do cLine[i] = ' '; while (i++ < 16);

		FormatI(&cLine[0], g_wSetI * 10);
		cLine[6] = ':';
		FormatI(&cLine[8], g_wGetI);

		if (g_nMode == MODE_NORMAL)
		{
			if (((s_wGetI + 5) / 10) >= s_wSetI) {
				cLine[15] = '\x9F';
			}
			else {
				cLine[15] = ' ';
			}
		}
		else {
			cLine[15] = g_nMode - 1 + '0';
		}
	
		DisplayText(0, 1, cLine);
		s_wSetI = g_wSetI;
		s_wGetI = g_wGetI;
		s_wMode = g_nMode;
	}
}

// ●電圧値テキスト生成
static void FormatV(char* pBuf, WORD wVal)
{
	// "12.00V"  wVal= mV
	if (wVal >= 10000) {
		pBuf[0] = (wVal / 10000) + '0';
		wVal %= 10000;
	}
	pBuf[1] = (wVal / 1000) + '0';
	wVal %= 1000;

	pBuf[2] = '.';

	pBuf[3] = (wVal / 100) + '0';
	wVal %= 100;

	pBuf[4] = (wVal / 10) + '0';

	pBuf[5] = 'V';
}

// ●電流値テキスト生成
static void FormatI(char* pBuf, WORD wVal)
{
	// "1000mA"  wVal = mA
	WORD wDsp = wVal;
	if (wVal >= 1000) {
		pBuf[0] = (wDsp / 1000) + '0';
		wDsp %= 1000;
	}
	if (wVal >= 100) {
		pBuf[1] = (wDsp / 100) + '0';
		wDsp %= 100;
	}
	if (wVal >= 10) {
		pBuf[2] = (wDsp / 10) + '0';
		wDsp %= 10;
	}
	pBuf[3] = wDsp + '0';
	
	pBuf[4] = 'm';
	pBuf[5] = 'A';
}

// ●文字列表示
static void DisplayText(int x, int y, const char* pText)
{
	BYTE data = 0x80; // SetDDRAM address
	data |= y << 6; // Line1/2 start address
	data += x;

	WriteLCD(0, data);
	for (; *pText != '\0'; pText++) {
		WriteLCD(1, *pText);
	}
}

// ●RE入力
static int GetRE(void)
{
	int nRet = GetRE1();
	return (nRet != RE_NONE)? nRet : GetRE2();
}

// ●RE1入力
static int GetRE1(void)
{
	BOOL bCurA = RE1_A;

	// A signal が立ち下がった時の B signal を調べる
	if (bCurA != g_bAsig1)
	{
		g_bAsig1 = bCurA;

		if (!bCurA) {
			return (RE1_B)? RE1_UP : RE1_DOWN;
		}
	}

	return RE_NONE;
}

// ●RE1入力
static int GetRE2(void)
{
	BOOL bCurA = RE2_A;

	// A signal が立ち下がった時の B signal を調べる
	if (bCurA != g_bAsig2)
	{
		g_bAsig2 = bCurA;

		if (!bCurA) {
			return (RE2_B)? RE2_UP : RE2_DOWN;
		}
	}

	return RE_NONE;
}

// ●操作処理
static void Operation(int nRE)
{
	if (g_nMode == MODE_NORMAL)
	{
		// 通常動作
		switch (nRE)
		{
		case RE1_UP:
			if (++g_wSetV > MAX_V) {
				g_wSetV = MAX_V;
			}
			else {
				WriteEEP(EEP_SETV, g_wSetV);
			}
			g_nMinV = 0;
			break;

		case RE1_DOWN:
			if (--g_wSetV < MIN_V) {
				g_wSetV = MIN_V;

				if ((++g_nMinV >= TO_TEST) & (g_nMinI >= TO_TEST)) {
					ToTest();
				}
			}
			else {
				WriteEEP(EEP_SETV, g_wSetV);
			}
			break;
		
		case RE2_UP:
			if (++g_wSetI > MAX_I) {
				g_wSetI = MAX_I;
			}
			else {
				WriteEEP(EEP_SETI, g_wSetI);
			}
			g_nMinI = 0;
			break;

		case RE2_DOWN:
			if (--g_wSetI < MIN_I) {
				g_wSetI = MIN_I;

				if ((++g_nMinI >= TO_TEST) & (g_nMinV >= TO_TEST)) {
					ToTest();
				}
			}
			else {
				WriteEEP(EEP_SETI, g_wSetI);
			}
			break;
		}
	}
	else
	{
		// テストモード
		switch (nRE)
		{
		case RE1_UP:
			// テストモード切替(UP)
			if (++g_nMode > MODE_TEST3) {
				g_nMode = MODE_TEST0;
			}
			g_wSetV = (g_nMode == MODE_TEST1)? TEST_V : 0;
			break;
			
		case RE1_DOWN:
			// テストモード切替(DOWN)
			if (--g_nMode < MODE_TEST0) {
				g_nMode = MODE_TEST3;
			}
			g_wSetV = (g_nMode == MODE_TEST1)? TEST_V : 0;
			break;

		case RE2_UP:
			switch (g_nMode)
			{
			// テストモード終了
			case MODE_TEST0:
				WriteEEP(EEP_FLAG, COMPLETE_VAL); // テスト完了にする
				ToNormal();
				break;

			// 自動校正開始
			case MODE_TEST2:
				AutoAdjV();
				break;

			// 自動校正開始
			case MODE_TEST3:
				AutoAdjI();
				break;
			}
			break;
		}
	}
}

// ●通常モードへ
static void ToNormal(void)
{
	g_nMode = MODE_NORMAL;
	g_wSetV = ReadEEP(EEP_SETV);
	g_wSetI = ReadEEP(EEP_SETI);

	g_wGetV = 0;
	g_wGetI = 0;
	
	Display();
	DelayMs(500);

	g_nMinV = 0;
	g_nMinI = 0;
}

// ●テストモードへ
static void ToTest(void)
{
	g_nMode = MODE_TEST0;
	g_wSetV = 0;
	g_wSetI = 0;

	g_wGetV = 0;
	g_wGetI = 0;

	Display();
	DelayMs(500);
}

// ●自動校正（電圧）
static void AutoAdjV(void)
{
	WORD wDAC = 1;
	g_wSetV = MIN_V;
	WORD wTest = g_wSetV * 100;
	WORD wMin = 0xFFFF;

	SetDAC(0xFFFF, ReadEEP(EEP_ISET_BASE + 10)); // 100mA

	while(g_wSetV <= MAX_V)
	{
		SetDAC(wDAC, 0xFFFF);
		DelayMs(10);
		Measurement();
		Display();
		
		WORD wDiff = (wTest >= g_wGetV)? (wTest - g_wGetV) : (g_wGetV - wTest);

		if ((wDiff < 20) && (wDiff > wMin))
		{
			WriteEEP(EEP_VSET_BASE + g_wSetV, wDAC);
			g_wSetV++;
			wTest += 100;
			wMin = 0xFFFF;
		}
		else if (wDiff < wMin) {
			wMin = wDiff;
		}
		if (++wDAC > 0xFFF) {
			wDAC = 1;
		}

		ClrWdt();
	}
	g_wSetV = 0;
	g_wGetV = 0;
	SetDAC(0, 0);
}

// ●自動校正（電流）
static void AutoAdjI(void)
{
	WORD wDAC = 1;
	g_wSetI = MIN_I;
	WORD wTest = g_wSetI * 10;
	
	SetDAC(ReadEEP(EEP_VSET_BASE + 50), 0xFFFF); // 5V

	while(g_wSetI <= MAX_I)
	{
		SetDAC(0xFFFF, wDAC);
		DelayMs(10);
		Measurement();
		Display();
		
		WORD wDiff = (wTest >= g_wGetI)? (wTest - g_wGetI) : (g_wGetI - wTest);

		if (wDiff == 0)
		{
			WriteEEP(EEP_ISET_BASE + g_wSetI, wDAC);
			g_wSetI++;
			wTest += 10;
		}
		if (++wDAC > 0xFFF) {
			wDAC = 1;
		}

		ClrWdt();
	}
	g_wSetI = 0;
	g_wGetI = 0;
	SetDAC(0, 0);
}

// ●EEPROM(DATA)へ読み出し
static WORD ReadEEP(WORD wAdd)
{
	wAdd *= 2;
	while(NVMCONbits.WR);

	TBLPAG = 0x7F;
	return __builtin_tblrdl(0xFE00 + wAdd);	
}

// ●EEPROM(DATA)へ書き込み
static void WriteEEP(WORD wAdd, WORD wData)
{
	wAdd *= 2;
	while(NVMCONbits.WR);

	TBLPAG = 0x7F;

	NVMCON = 0x4058;
	__builtin_tblwtl(0xFE00 + wAdd, 0xFFFF);

	asm volatile ("disi #5");
	__builtin_write_NVM();

	while(NVMCONbits.WR);

	NVMCON = 0x4004;
	__builtin_tblwtl(0xFE00 + wAdd, wData);

	asm volatile ("disi #5");
    __builtin_write_NVM();
}

// ●LCD初期化(8Bit mode)
static void InitOLED(void)
{
	WriteLCD(0, 0b00111001); // Function Set (FT:01 西欧キャラクタフォントＩ)
	WriteLCD(0, 0b00111001); // Function Set
	WriteLCD(0, 0b00111001); // Function Set
	WriteLCD(0, 0b00111001); // Function Set

	WriteLCD(0, 0b00001000); // Display Off
	WriteLCD(0, 0b00000001); // Display Clear
	DelayMs(7);
	WriteLCD(0, 0b00000110); // Entry Mode Set
	
	WriteLCD(0, 0b00001100); // Display On
}

// ●LCDへ書き込み
static void WriteLCD(BYTE rs, BYTE data)
{
	LCD_RS = rs;
	//DelayUs(20); // tAS68:20us

	LCD_DB0 = (data & 0x01)? 1 : 0;
	LCD_DB1 = (data & 0x02)? 1 : 0;
	LCD_DB2 = (data & 0x04)? 1 : 0;
	LCD_DB3 = (data & 0x08)? 1 : 0;
	LCD_DB4 = (data & 0x10)? 1 : 0;
	LCD_DB5 = (data & 0x20)? 1 : 0;
	LCD_DB6 = (data & 0x40)? 1 : 0;
	LCD_DB7 = (data & 0x80)? 1 : 0;

	LCD_E = 1;
	DelayUs(250); // tPW68(W):250us

	LCD_E = 0;
	DelayUs(20); // tDH68:20us

	//DelayUs(230); // tCY68:500us
}

// ●DAC(MCP4922)へ出力する
static void SetDAC(WORD wA, WORD wB)
{
	// A(電圧設定)
	if (wA != 0xFFFF)
	{
		DAC_CS = 0;
		wA |= 0x3000; // Write to DACa BUF=Unbufferd GA=1x SHDN=1
		SendSPI(wA >> 8);
		SendSPI(wA & 0xFF);
		DAC_CS = 1;
	}

	// B(電流設定)
	if (wB != 0xFFFF)
	{
		DAC_CS = 0;
		wB |= 0xB000; // Write to DACb BUF=Unbufferd GA=1x SHDN=1
		SendSPI(wB >> 8);
		SendSPI(wB & 0xFF);
		DAC_CS = 1;
	}
}

// ●SPI送受信
static BYTE SendSPI(BYTE bData)
{
	SPI2BUF = bData;
	while(!SPI2STATbits.SPIRBF);
	return (BYTE)SPI2BUF;
}

// ●ADC値を取得する
static WORD GetADC(BYTE ch)
{
	_CH0SA = ch; // ADC-V:AN13 | ADC-I:AN14 | VBG:26

	for(;;) // たまにDONEがセットされないことがあるのでリトライ機構を入れる
	{
		AD1CON1bits.SAMP = 1;
		int i = 0;
		do{
			if (AD1CON1bits.DONE) {
				break;
			}
		} while (++i < 15); // 10TADで終わるはず
		if (i < 15) {
			break;
		}
	}

	return ADC1BUF0;
}

// ●ウェイト(MS)
static void DelayMs(WORD wMs)
{
	PR2 = 4000 - 18;
	do {
		TMR2 = 0;
		_T2IF = 0;
		while(!_T2IF);
	} while(--wMs > 0);
}

// ●ウェイト(US)
static void DelayUs(WORD wUs)
{
	wUs /= 10;
	PR2 = 40 - 14;
	do {
		TMR2 = 0;
		_T2IF = 0;
		while(!_T2IF);
	} while(--wUs > 0);
}
